/* Copyright (C) 2005,2006 G.P. Halkes
   Licensed under the Open Software License version 2.0 */

#ifndef GLOBALS_H 
#define GLOBALS_H 

/* This should be automatically determined from the repository
   [and is, by the mkdist.sh script] */
#define VERSION_STRING "0.3.1"
/* Version number in HEX. Must be defined to current version to allow self
   compilation. Will be set to correct value by mkdist.sh script. Used to
   create a define in the output such that users can #ifdef for features. */
#define VERSION_HEX "0x000301L"

#include <stdio.h>
#include "nonRuleAnalysis.h"

/* Variables */
extern bool errorSeen, softErrorSeen;

/* Functions */

/* Allocate a block of memory, and die if it cannot be allocated */
void *safeMalloc(size_t size, const char *where);
/* Is the same as safeMalloc, but also clears the memory block */
void *safeCalloc(size_t size, const char *where);
void *safeRealloc(void *ptr, size_t size, const char *where);
char *safeStrdup(const char *orig, const char *where);
char *safeStrndup(const char *orig, size_t length, const char *where);
FILE *safeFopen(const char *name, const char *mode);

#define WARNING_UNMASKED (0)
#define WARNING_ARG_SEPARATOR (1<<0)
#define WARNING_OPTION_OVERRIDE (1<<1)
#define WARNING_UNBALANCED_C (1<<2)
#define WARNING_MULTIPLE_PARSER (1<<3)
#define WARNING_EOFILE (1<<4)
#define WARNING_UNUSED (1<<5)
#define WARNING_DATATYPE (1<<6)

/* Documentation for the functions below:
fatal:   To be used internally when a something goes horribly wrong
error:   To be used when an error in the syntax or semantics is detected
warning: To be used for warnings about the syntax or semantics
*/
/* If gcc is used, it can check whether the correct types are used for the
   arguments after the format string. */
#ifdef __GNUC__
void fatal(const char *fmt, ...) __attribute__((format (printf, 1, 2))) __attribute__((noreturn));
void error(Token *token, const char *fmt, ...) __attribute__((format (printf, 2, 3)));
void softError(Token *token, const char *fmt, ...) __attribute__((format (printf, 2, 3)));
void warning(int warningType, Token *token, const char *fmt, ...) __attribute__((format (printf, 3, 4)));
void continueMessage(const char *fmt, ...)  __attribute__((format (printf, 1, 2)));
#else
/*@noreturn@*/void fatal(const char *fmt, ...);
void error(Token *token, const char *fmt, ...);
void softError(Token *token, const char *fmt, ...);
void warning(int warningType, Token *token, const char *fmt, ...);
void continueMessage(const char *fmt, ...);
#endif
void printAt(Token *token);
void endMessage(void);

#define PANIC() fatal("Program failure in file %s on line %d. Please see Section Bugs in the manual.\n", __FILE__, __LINE__)
#ifdef DEBUG
#define PANICMSG(msg) fatal("%s:%d: %s", __FILE__, __LINE__, msg)
#else
#define PANICMSG(msg) PANIC()
#endif
#define ASSERT(cond) if (!(cond)) PANIC()

char *processString(Token *file);
char *strtokReentrant(char *string, const char *separators, char **state);

/* Don't want to bother making a header file just for this. */
void freeMemory(void);

#define uDeclaration un.declaration
#define uDirective un.directive
#define uCode un.code
#define uToken un.token
#define uLiteral un.literal
#define uNumber un.number
#define uNonTerminal un.nonTerminal
#define uTerminal un.terminal
#define uTerm un.term
#define uAction un.action
#define uNext un.next
#define uPtr un.ptr

#endif

