/* Copyright (C) 2005,2006 G.P. Halkes
   Licensed under the Open Software License version 2.0 */

%options "generate-symbol-table lowercase-symbols reentrant";
%lexical lexerWrapper;
{  /* start of C code */
#include <errno.h>
#include <stdlib.h>
#include <limits.h>

#include "posixregex.h"

#include "nonRuleAnalysis.h"
#include "ruleAnalysis.h"
#include "globals.h"
#include "argcount.h"
#include "option.h"
#include "localOptionMacros.h"
#include "lexer.h"
#include "bool.h"

/** Function required by parser to show error messages.
	@param symb A symbol to be inserted or a value indicating the deletion of
		tokens.
*/
void LLmessage(int symb) {
	switch (symb) {
		case LL_MISSINGEOF:
			error(NULL, "Unexpected %s", LLgetSymbol(LLsymb));
			break;
		case LL_DELETE:
			error(NULL, "Unexpected %s", LLgetSymbol(LLsymb));
			break;
		default:
			error(NULL, "Expected %s before %s", LLgetSymbol(symb), LLgetSymbol(LLsymb));
			break;
	}
	switch (LLsymb) {
		case IDENTIFIER:
		case NUMBER:
		case LITERAL:
		case STRING:
			continueMessage(" %s", yytext);
			break;
		default:
			break;
	}
	endMessage();
}

} /* end of C code */
%token IDENTIFIER, NUMBER, LITERAL, STRING;

%label MISSING_ELEMENT, "literal, identifier or term";
%label MISSING_LITERAL_OR_IDENTIFIER, "literal or identifier";
%label MISSING_STRING, "string";
%label MISSING_IDENTIFIER, "identifier";

%label TOKEN, "%token";
%label START, "%start";
%label PERSISTENT, "%persistent";
%label IF, "%if";
%label WHILE, "%while";
%label AVOID, "%avoid";
%label PREFER, "%prefer";
%label DEFAULT, "%default";
%label LEXICAL, "%lexical";
%label PREFIX, "%prefix";
%label ONERROR, "%onerror";
%label FIRST, "%first";
%label C_DECL, "C code";
%label C_EXPR, "C expression";
%label LABEL, "%label";
%label INCLUDE, "%include";
%label OPTIONS, "%options";
%label DATATYPE, "%datatype";

%start parser, specification;

specification :
	[ declaration ]*
;

identifier(Token **token) :
	IDENTIFIER
	{
		*token = newToken();
	}
|
	%default
	MISSING_IDENTIFIER
;

declaration 
{
	Token *token = NULL, *secondToken = NULL;
} :
	C_DECL
	{
		CCode *code = newCCode();
		newDeclaration(CODE, code);
	}
|
	START
	identifier(&token)
	','
	identifier(&secondToken)
	{
		if (token != NULL && secondToken != NULL) {
			Directive *directive = newDirective(START_DIRECTIVE, token, secondToken);
			newDeclaration(DIRECTIVE, directive);
		} else {
			freeToken(token);
			freeToken(secondToken);
		}
	}
	';'
|
	TOKEN
	identifier(&token)
	{
		if (token != NULL) {
			Directive *directive = newDirective(TOKEN_DIRECTIVE, token, NULL);
			newDeclaration(DIRECTIVE, directive);
			token = NULL;
		}
	}
	[
		','
		identifier(&token)
		{
			if (token != NULL) {
				Directive *directive = newDirective(TOKEN_DIRECTIVE, token, NULL);
				newDeclaration(DIRECTIVE, directive);
				token = NULL;
			}
		}
	]*
	';'
|
	FIRST
	identifier(&token)
	','
	identifier(&secondToken)
	{
		if (token != NULL && secondToken != NULL) {
			Directive *directive = newDirective(FIRST_DIRECTIVE, token, secondToken);
			newDeclaration(DIRECTIVE, directive);
		} else {
			freeToken(token);
			freeToken(secondToken);
		}
	}
	';'
|
	LEXICAL
	identifier(&token)
	{
		if (token != NULL) {
			Directive *directive = newDirective(LEXICAL_DIRECTIVE, token, NULL);
			newDeclaration(DIRECTIVE, directive);
		}
	}
	';'
|
	PREFIX
	identifier(&token)
	{
		if (token != NULL) {
			Directive *directive = newDirective(PREFIX_DIRECTIVE, token, NULL);
			newDeclaration(DIRECTIVE, directive);
		}
	}
	';'
|
	ONERROR
	identifier(&token)
	{
		if (token != NULL) {
			Directive *directive = newDirective(ONERROR_DIRECTIVE, token, NULL);
			newDeclaration(DIRECTIVE, directive);
		}
		error(NULL, "%%onerror not supported (yet)\n");
	}
	';'
|
	LABEL
	[
		IDENTIFIER
		{
			if (!option.llgenMode)
				token = newToken();
		}
	|
		LITERAL
		{
			if (!option.llgenMode)
				token = newToken();
		}
	|
		%default
		MISSING_LITERAL_OR_IDENTIFIER
	]
	','
	[
		STRING
		{
			if (!option.llgenMode)
				secondToken = newToken();
		}
	|
		%default
		MISSING_STRING
	]
	{
		if (option.llgenMode) {
			error(NULL, "%%label is not allowed in LLgen mode\n");
			freeToken(token);
			freeToken(secondToken);
		} else if (token != NULL && secondToken != NULL) {
			Directive *directive = newDirective(LABEL_DIRECTIVE, token, secondToken);
			newDeclaration(DIRECTIVE, directive);
		} else {
			freeToken(token);
			freeToken(secondToken);
		}
	}
	';'
|
	INCLUDE
	[
		STRING
		{
			if (!(option.llgenMode || option.LLgenStyleOutputs)) {
				token = newToken();
			}
		}
	|
		%default
		MISSING_STRING /* token is NULL by default */
	]
	';'
	{
		if (option.llgenMode) {
			error(NULL, "%%include is not allowed in LLgen mode\n");
		} else if (option.LLgenStyleOutputs) {
			error(NULL, "%%include is not allowed with LLgen-style outputs\n");
		} else if (token != NULL) {
			if (openInclude(token))
				parser();
			freeToken(token);
		}
	}
|
	OPTIONS
	[
		STRING
		{
			if (!option.llgenMode)
				token = newToken();
		}
	|
		%default
		MISSING_STRING /* token is NULL by default */
	]
	';'
	{
		
		/* Most options will override the command line, except for:
			--base-name
			--error-limit
			--verbose
			--token-pattern
		*/
		if (option.llgenMode) {
			error(NULL, "%%options is not allowed in LLgen mode\n");
		} else if (token != NULL) {
			/* NOTE: this name is fixed, as all the macro's below use it! */
			char *optstring;
			
			char *strtokState;
			char *processedString = processString(token);
			
			if (processedString == NULL) {
				warning(WARNING_UNMASKED, token, "Error parsing %%options string\n");
			} else if ((optstring = strtokReentrant(processedString, " \t\n", &strtokState)) != NULL) {
				bool nonSupported = false;
				do {
					PREPROCESS_OPTION()

					LOCAL_OPTION("verbose", OPTIONAL_ARG)
						if (!option.verboseSet) {
							if (optarg != NULL) {
								PARSE_INT(option.verbose, 1, MAX_VERBOSITY)
							} else { 
								option.verbose++;
								/* Prevent overflow */
								if (option.verbose > MAX_VERBOSITY)
									option.verbose = MAX_VERBOSITY;
							}
						} else {
							warning(WARNING_OPTION_OVERRIDE, token, "Will not override verbose option passed on command line\n");
						}
					LOCAL_END_OPTION
					LOCAL_OPTION("suppress-warnings", OPTIONAL_ARG)
						checkWarningSuppressionArguments(optarg, (int) optlength, optstring);
					LOCAL_END_OPTION
				
					LOCAL_OPTION("max-compatibility", NO_ARG)
						option.LLgenArgStyle = true;
						option.onlyLLgenEscapes = true;
						option.LLgenStyleOutputs = true;
					LOCAL_END_OPTION
					LOCAL_BOOLEAN_OPTION("warnings-as-errors", option.warningsErrors)
					LOCAL_OPTION("base-name", REQUIRED_ARG)
						if (!option.outputBaseNameSet)
							option.outputBaseName = safeStrdup(optarg, "parseCmdLine");
						else
							warning(WARNING_OPTION_OVERRIDE, token, "Will not override base-name option passed on command line\n");
					LOCAL_END_OPTION
					LOCAL_BOOLEAN_OPTION("llgen-arg-style", option.LLgenArgStyle)
					LOCAL_BOOLEAN_OPTION("llgen-escapes-only", option.onlyLLgenEscapes)
					LOCAL_BOOLEAN_OPTION("no-prototypes-header", option.noPrototypesHeader)
					LOCAL_BOOLEAN_OPTION("no-line-directives", option.dontGenerateLineDirectives)
					LOCAL_BOOLEAN_OPTION("llgen-output-style", option.LLgenStyleOutputs)
					LOCAL_BOOLEAN_OPTION("reentrant", option.reentrant)
					LOCAL_BOOLEAN_OPTION("no-eof-zero", option.noEOFZero)
					LOCAL_BOOLEAN_OPTION("no-llreissue", option.noLLreissue)
					LOCAL_BOOLEAN_OPTION("generate-lexer-wrapper", option.generateLexerWrapper)
					LOCAL_BOOLEAN_OPTION("generate-llmessage", option.generateLLmessage)
					LOCAL_BOOLEAN_OPTION("generate-symbol-table", option.generateSymbolTable)
					LOCAL_BOOLEAN_OPTION("abort", option.abort)
					LOCAL_BOOLEAN_OPTION("no-allow-label-create", option.noAllowLabelCreate)
					LOCAL_OPTION("error-limit", REQUIRED_ARG)
						if (!option.errorLimitSet)
							PARSE_INT(option.errorLimit, 0, INT_MAX)
						else
							warning(WARNING_OPTION_OVERRIDE, token, "Will not override error-limit option passed on the command line\n");
					LOCAL_END_OPTION
					LOCAL_BOOLEAN_OPTION("no-arg-count", option.noArgCount);
					LOCAL_BOOLEAN_OPTION("lowercase-symbols", option.lowercaseSymbols);
					LOCAL_BOOLEAN_OPTION("not-only-reachable", option.notOnlyReachable);
					LOCAL_BOOLEAN_OPTION("thread-safe", option.threadSafe)
#ifdef USE_REGEX
					LOCAL_OPTION("token-pattern", REQUIRED_ARG)
						int result;
						if (!option.useTokenPattern) {
							option.useTokenPattern = true;
							if ((result = regcomp(&option.tokenPattern, optarg, REG_EXTENDED | REG_NOSUB)) != 0) {
								char errorBuffer[100];
								regerror(result, &option.tokenPattern, errorBuffer, 100);
								fatal("Regular expression could not be compiled: %s\n", errorBuffer);
							}
						} else {
							warning(WARNING_OPTION_OVERRIDE, token, "Will not override token-pattern specified on the command line\n");
						}
					LOCAL_END_OPTION
#else
					LOCAL_OPTION("token-pattern", OPTIONAL_ARG)
						fatal("This LLnextgen binary does not support the '%.*s' option. See the documentation for details\n", (int) optlength, optstring);
					LOCAL_END_OPTION
#endif
					/* Show appropriate message for options only supported on command line */
					LOCAL_OPTION_USED("depend", nonSupported)
					LOCAL_OPTION_USED("depend-cpp", nonSupported)
					LOCAL_OPTION_USED("dump-lexer-wrapper", nonSupported)
					LOCAL_OPTION_USED("dump-llmessage", nonSupported)
					LOCAL_OPTION_USED("dump-tokens", nonSupported)
					LOCAL_OPTION_USED("help", nonSupported)
					LOCAL_OPTION_USED("keep-dir", nonSupported)
					LOCAL_OPTION_USED("show-dir", nonSupported)
					LOCAL_OPTION_USED("version", nonSupported)

					if (nonSupported)
						fatal("Option '%.*s' is not supported in %%options\n", (int) optlength, optstring);

					error(token, "Option '%.*s' does not exist\n", (int) optlength, optstring);
				} while ((optstring = strtokReentrant(NULL, " \t\n", &strtokState)) != NULL);
				
				/* After this point, the string will not be used anymore. */
				free(processedString);

				postOptionChecks();
			}
			/* After this point, the token will not be used anymore. */
			freeToken(token);
		}
	}
|
	DATATYPE
	[
		STRING
		{
			if (!option.llgenMode)
				token = newToken();
		}
	|
		%default
		MISSING_STRING
	]
	[
		','
		[
			STRING
			{
				if (!option.llgenMode)
					secondToken = newToken();
			}
		|
			%default
			MISSING_STRING
		]
	]?
	';'
	{
		if (option.llgenMode) {
			error(NULL, "%%datatype is not allowed in LLgen mode\n");
		} else {
			Directive *directive = newDirective(DATATYPE_DIRECTIVE, token, secondToken);
			newDeclaration(DIRECTIVE, directive);
		}
	}
|
	rule ';'
;

rule
{
	NonTerminal *nonTerminal;
	Token *name = NULL;
	CCode *expr = NULL, *decl = NULL;
	List *rule;
	int argCount = 0;
} :
	identifier(&name) /* Name */
	[
		C_EXPR
		{
			expr = newCCode();
			/* The input may not be completely valid, but we can still do
			   checking without worying about seg-faults and the like. We may
			   generate an extra error message, but that's OK. */
			if (!option.noArgCount)
				argCount = determineArgumentCount(expr, true);
		}
	]?		/* Parameters */
	[
		C_DECL
		{
			decl = newCCode();
		}
	]?		/* Local variable definitions */
	':'
	{
		rule = newList();
		nonTerminal = newNonTerminal(name, expr, decl, rule, argCount);
		if (name != NULL)
			newDeclaration(NONTERMINAL, nonTerminal);
	}
	productions(&nonTerminal->term)
;

productions(Term *term)
{
	Alternative *alternative;
} :
	{
		alternative = newAlternative();
		alternative->enclosing = term;
		listAppend(term->rule, alternative);
	}
	simpleproduction(alternative)
	[
		'|'
		{
			alternative = newAlternative();
			alternative->enclosing = term;
			listAppend(term->rule, alternative);
		}
		simpleproduction(alternative)
	]*
;

simpleproduction(Alternative *alternative) 
{
	GrammarPart *grammarPart;
} :
	[
		DEFAULT
		{
			if (alternative->flags & ALTERNATIVE_DEFAULT)
				error(NULL, "%%default specified more than once on a single alternative\n");
			alternative->flags |= ALTERNATIVE_DEFAULT;
		}
	|
		IF
		{
			if (alternative->flags & ALTERNATIVE_CONDITION)
				error(NULL, "More than one condition specified on a single alternative\n");
			alternative->flags |= ALTERNATIVE_IF;
		}
		C_EXPR
		{
			alternative->expression = newCCode();
		}
	|
		PREFER
		{
			if (alternative->flags & ALTERNATIVE_CONDITION)
				error(NULL, "More than one condition specified on a single alternative\n");
			alternative->flags |= ALTERNATIVE_PREFER;
		}
	|
		AVOID
		{
			if (alternative->flags & ALTERNATIVE_CONDITION)
				error(NULL, "More than one condition specified on a single alternative\n");
			alternative->flags |= ALTERNATIVE_AVOID;
		}
	]*
	[
		%persistent
		element(&grammarPart)
		repeats(&(grammarPart->repeats))
		{
			switch (grammarPart->subtype) {
				case PART_ACTION:
					if (grammarPart->repeats.number < 0 || grammarPart->repeats.number > 1 || (grammarPart->repeats.subtype & STAR))
						error(grammarPart->uAction, "Repetition specified on action\n");
					break;
				case PART_UNDETERMINED:
				case PART_LITERAL:
					if (grammarPart->repeats.subtype & FIXED && grammarPart->repeats.number == 1)
						break;

					/* This will make it a Term, so treat it like one! */
					grammarPart = wrapInTerm(grammarPart);
					/* FALLTHROUGH */
				case PART_TERM:
					/* I'm not merging in terms with a single alternative, because I want to
					   keep the information about conflict resolvers. Then we can bug the user ;-) */
					grammarPart->uTerm.term.index = listSize(alternative->parts);
					grammarPart->uTerm.term.enclosing.alternative = alternative;
					break;
				case PART_UNKNOWN:
					break;
				default:
					PANIC();
			}
			listAppend(alternative->parts, grammarPart);
		}
	]*
;

element(GrammarPart **part) :
	%default
	MISSING_ELEMENT
	{
		*part = newGrammarPart(NULL, PART_UNKNOWN);
	}
|
	C_DECL	/* Action */
	{
		*part = newGrammarPart(NULL, PART_ACTION);
		(*part)->uAction = newCCode();
	}
|
	'['
	{
		*part = newGrammarPart(newToken(), PART_TERM);
	}
	[
		WHILE
		C_EXPR
		{
			if ((*part)->uTerm.term.flags & TERM_WHILE) {
				error(NULL, "%%while specified more than once in a single term\n");
			} else {
				(*part)->uTerm.term.flags |= TERM_WHILE;
				(*part)->uTerm.expression = newCCode();
			}
		}
	|
		PERSISTENT
		{
			if ((*part)->uTerm.term.flags & TERM_PERSISTENT)
				error(NULL, "%%persistent specified more than once in a single term\n");
			(*part)->uTerm.term.flags |= TERM_PERSISTENT;
		}
	]*
	/* List is already initialised as part of newGrammarPart */
	productions(&(*part)->uTerm.term)
	']'
|
	LITERAL
	{
		*part = newGrammarPart(newToken(), PART_LITERAL);
		/* GRAM_TERMINAL(*part) = literalToTerminal((*part)->token->text); */
	}
|
	IDENTIFIER /* Both terminal and non-terminal */
	{
		*part = newGrammarPart(newToken(), PART_UNDETERMINED);
	}
	[
		C_EXPR
		{
			(*part)->uNonTerminal.expression = newCCode();
		}
	]? /* Parameters */
;

repeats(Repeats *reps) :
	/* No repetition operator specified, so exactly one match is requested */
	{
		reps->subtype = FIXED;
		reps->number = 1;
	}
|
	[
		'*'
		{
			reps->subtype = STAR;
		}
		
	|
		'+'
		{
			reps->subtype = PLUS;
		}
	]
	[
		NUMBER
		{
			long value;
			errno = 0;
			value = strtol(yytext, NULL, 10);
			if (errno == ERANGE || value > INT_MAX)
				error(NULL, "Number of repetitions too large\n");
			if (value <= 0) {
				PANIC(); 
				/* The lexer only returns numbers without a sign, therefore the
				   parser should PANIC when it does happen. */
			}
			reps->number = value;
			if (reps->subtype == PLUS && reps->number == 1) {
				reps->subtype = FIXED;
				warning(WARNING_UNMASKED, NULL, "Specifying fixed single repetition as '+1'\n");
			}
		}
	]?
|
	NUMBER
	{
		long value;
		reps->subtype = FIXED;
		errno = 0;
		value = strtol(yytext, NULL, 10);
		if (errno == ERANGE || value > INT_MAX)
			error(NULL, "Number of repetitions too large\n");
		if (value <= 0) {
			PANIC(); 
			/* The lexer only returns numbers without a sign, therefore the
			   parser should PANIC when it does happen. */
		}
		reps->number = value;
	}
|
	'?'
	{
		reps->subtype = STAR;
		reps->number = 1;
	}
;


