/* Copyright (C) 2005,2006 G.P. Halkes
   Licensed under the Open Software License version 2.0 */

#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <limits.h>

#include "posixregex.h"

#include "optionMacros.h"
#include "option.h"
#include "globals.h"
#include "os.h"
#include "generate.h"

options option;

const char *descriptions[] = {
"    -c, --max-compatibility      Set options required for maximum compatibility\n",
"                                 This implies --llgen-arg-style,\n"
"                                 --llgen-escapes-only and --llgen-output-style\n"
"    -e, --warnings-as-errors     Treat warnings as errors\n",
"    -Enum, --error-limit=num     Set the maximum number of errors\n",
"    -h, --help                   Print out this help message\n",
"    -V, --version                Print the program version and exit\n",
"    -v[level], --verbose[=level] Increase or set the verbosity level\n",
"    -w[warnings], --suppress-warnings[=warnings]\n",
"                                 Suppress all or selected warnings. Available\n",
"                                 warnings are: arg-separator, option-override,\n"
"                                 unbalanced-c, multiple-parser, eofile,\n",
"                                 unused[:<identifier>] and datatype\n",
"    --abort                      Generate the LLabort function\n",
"    --base-name=name             Set the base name for the output files\n",
"    --depend[=modifiers]         Generate dependency information. Modifiers can\n",
"                                 be used to change what is output. Available\n",
"                                 modifiers are: targets:<targets>, file:<file>\n",
"                                 and extra-targets:<targets>\n",
"    --depend-cpp                 Extract code to pass through the\n",
"                                 C preproccessor for dependency information\n",
"    --dump-lexer-wrapper         Dump the lexer wrapper function to screen\n",
"    --dump-llmessage             Dump the default LLmessage function to screen\n",
#ifdef USE_REGEX
"    --dump-tokens[=modifier]     Dump %token directives for unknown identifiers\n",
"                                 that match the --token-pattern pattern\n",
"                                 Available modifiers: separate, labels\n",
#endif
"    --generate-lexer-wrapper     Generate a wrapper for the lexical analyser\n",
"    --generate-llmessage         Generate an LLmessage function\n",
"    --generate-symbol-table      Generate a symbol table\n",
"    --keep-dir                   Create outputs in same directory as inputs\n",
"    --llgen-arg-style            Use semicolons as argument separator\n",
"    --llgen-escapes-only         Only use LLgen character-literal escapes\n",
"    --llgen-output-style         Generate Lpars.[ch] and one .c file per input\n",
"    --lowercase-symbols          Convert names to lowercase in symbol table\n"
"    --no-allow-label-create      %label won't to create new tokens\n",
"    --no-arg-count               Do not check argument counts\n",
"    --no-eof-zero                Do not use 0 as end-of-file token\n",
"    --no-line-directives         Do not generate #line directives in the output\n",
"    --no-llreissue               Do not generate the LLreissue variable\n",
"    --no-prototypes-header       Do not generate a prototypes in the .h file\n",
"    --not-only-reachable         Do not only analyse reachable rules\n",
"    --reentrant                  Generate a reentrant parser\n",
"    --show-dir                   Show directory names in messages\n",
"    --thread-safe                Generate a thread safe parser\n",
#ifdef USE_REGEX
"    --token-pattern=pattern      Pattern for matching unknown identifiers\n",
#endif
NULL};

static void checkDependArguments(const char *optarg);

/** Parse command line arguments.
	@param argc The number of arguments on the command line.
	@param argv The array of character pointers containing the command line
		arguments.
*/
PARSE_FUNCTION(parseCmdLine)
	memset(&option, 0, sizeof(option));
	option.inputFileList = newList();
	option.errorLimit = -1;

	/* Check whether the program was called through the LLgen symlink */
	if (isLLgen(argv[0])) {
		option.llgenMode = true;
		option.LLgenArgStyle = true;
		option.onlyLLgenEscapes = true;
		option.LLgenStyleOutputs = true;
		option.noLLreissue = true;
		option.reentrant = true;
	}

	/* In LLgen mode we should only accept the original LLgen options. In
	   LLnextgen mode we can remove the original options. The -h option is 
	   then the same as --help in LLnextgen mode */
	OPTIONS
		/* ============= Options from LLgen ============= */
		if (option.llgenMode) {
			SHORT_OPTION('a', NO_ARG) /* LLgen: Generate ANSI compatible output. Ignored */
			END_OPTION
			SHORT_OPTION('h', REQUIRED_ARG) /* LLgen: Set high_percentage for jump table. Ignored */
				int discard;
				PARSE_INT(discard, 0, 100)
			END_OPTION
			SHORT_OPTION('j', OPTIONAL_ARG) /* LLgen: Generate dense switches. Ignored */
				if (optarg != NULL) {
					int discard;
					PARSE_INT(discard, 0, 100)
				}
			END_OPTION
			SHORT_OPTION('l', REQUIRED_ARG) /* LLgen: Set low_percentage for jump table. Ignored */
				int discard;
				PARSE_INT(discard, 0, 100)
			END_OPTION
			SHORT_OPTION('n', OPTIONAL_ARG);
				fatal("LLnextgen cannot generate parsers with non-correcting error-recovery\n");
			END_OPTION
			SHORT_OPTION('s', OPTIONAL_ARG);
				fatal("LLnextgen cannot generate parsers with non-correcting error-recovery\n");
			END_OPTION
			SHORT_OPTION('x', NO_ARG) /* LLgen: Generate extended sets. Ignored */
				fprintf(stderr, "LLnextgen cannot generate extended sets\n");
				break;
			END_OPTION
			SHORT_OPTION('v', NO_ARG)
				option.verbose++;
				/* Prevent overflow */
				if (option.verbose > MAX_VERBOSITY)
					option.verbose = MAX_VERBOSITY;
			END_OPTION
			BOOLEAN_SHORT_OPTION('w', option.suppressWarnings)
		} else {
		
			OPTION('v', "verbose", OPTIONAL_ARG)
				option.verboseSet = true;
				if (optarg != NULL) {
					PARSE_INT(option.verbose, 1, MAX_VERBOSITY)
				} else { 
					option.verbose++;
					/* Prevent overflow */
					if (option.verbose > MAX_VERBOSITY)
						option.verbose = MAX_VERBOSITY;
				}
			END_OPTION
			OPTION('w', "suppress-warnings", OPTIONAL_ARG)
				checkWarningSuppressionArguments(optarg, (int) optlength, CURRENT_OPTION);
			END_OPTION
			/* ============= New options ============= */
			OPTION('V', "version", NO_ARG)
				fputs("LLnextgen " VERSION_STRING "\nCopyright (C) 2005,2006 G.P. Halkes\nLicensed under the Open Software License version 2.0\n", stdout);
				exit(0);
			END_OPTION
			OPTION('c', "max-compatibility", NO_ARG)
				option.LLgenArgStyle = true;
				option.onlyLLgenEscapes = true;
				option.LLgenStyleOutputs = true;
			END_OPTION
			BOOLEAN_OPTION('e', "warnings-as-errors", option.warningsErrors)
			LONG_OPTION("base-name", REQUIRED_ARG)
				option.outputBaseNameSet = true;
				option.outputBaseName = safeStrdup(optarg, "parseCmdLine");
			END_OPTION
			BOOLEAN_LONG_OPTION("llgen-arg-style", option.LLgenArgStyle)
			BOOLEAN_LONG_OPTION("llgen-escapes-only", option.onlyLLgenEscapes)
			BOOLEAN_LONG_OPTION("keep-dir", option.keepDirInFilename)
			BOOLEAN_LONG_OPTION("no-prototypes-header", option.noPrototypesHeader)
			BOOLEAN_LONG_OPTION("no-line-directives", option.dontGenerateLineDirectives)
			BOOLEAN_LONG_OPTION("llgen-output-style", option.LLgenStyleOutputs)
			BOOLEAN_LONG_OPTION("reentrant", option.reentrant)
			BOOLEAN_LONG_OPTION("no-eof-zero", option.noEOFZero)
			BOOLEAN_LONG_OPTION("no-llreissue", option.noLLreissue)
			BOOLEAN_LONG_OPTION("generate-lexer-wrapper", option.generateLexerWrapper)
			BOOLEAN_LONG_OPTION("generate-llmessage", option.generateLLmessage)
			BOOLEAN_LONG_OPTION("generate-symbol-table", option.generateSymbolTable)
			BOOLEAN_LONG_OPTION("abort", option.abort)
			BOOLEAN_LONG_OPTION("no-allow-label-create", option.noAllowLabelCreate)
			BOOLEAN_LONG_OPTION("dump-lexer-wrapper", option.dumpLexerwrapper)
			BOOLEAN_LONG_OPTION("dump-llmessage", option.dumpLLmessage);
			OPTION('h', "help", NO_ARG)
				int i = 0;
				printf("Usage: LLnextgen [OPTIONS] [FILES]\n");
				while (descriptions[i] != NULL)
					fputs(descriptions[i++], stdout);
				exit(EXIT_SUCCESS);
			END_OPTION
			OPTION('E', "error-limit", REQUIRED_ARG)
				option.errorLimitSet = true;
				PARSE_INT(option.errorLimit, 0, INT_MAX)
			END_OPTION
			BOOLEAN_LONG_OPTION("show-dir", option.showDir);
			BOOLEAN_LONG_OPTION("no-arg-count", option.noArgCount);
			BOOLEAN_LONG_OPTION("lowercase-symbols", option.lowercaseSymbols);
			BOOLEAN_LONG_OPTION("not-only-reachable", option.notOnlyReachable);
			LONG_OPTION("depend", OPTIONAL_ARG)
				option.depend = true;
				checkDependArguments(optarg);
			END_OPTION
			BOOLEAN_LONG_OPTION("depend-cpp", option.dependCpp)
			BOOLEAN_LONG_OPTION("thread-safe", option.threadSafe)
#ifdef USE_REGEX
			LONG_OPTION("token-pattern", REQUIRED_ARG)
				int result;

				option.useTokenPattern = true;
				if ((result = regcomp(&option.tokenPattern, optarg, REG_EXTENDED | REG_NOSUB)) != 0) {
					char errorBuffer[100];
					regerror(result, &option.tokenPattern, errorBuffer, 100);
					fatal("Regular expression could not be compiled: %s\n", errorBuffer);
				}
			END_OPTION
			LONG_OPTION("dump-tokens", OPTIONAL_ARG)
				option.dumpTokens = true;
				if (optarg != NULL) {
					if (strcmp(optarg, "separate") == 0)
						option.dumpTokensType = DUMP_TOKENS_SEPARATE;
					else if (strcmp(optarg, "labels") == 0)
						option.dumpTokensType = DUMP_TOKENS_LABELS;
					/* Next is the default, so we don't include it in the docs,
					   but we allow it anyway. */
					else if (strcmp(optarg, "multiple") == 0)
						option.dumpTokensType = DUMP_TOKENS_MULTI;
					else
						fatal("Argument '%s' is invalid for option --depend\n", optarg);
				}
			END_OPTION
#else
			LONG_OPTION("token-pattern", OPTIONAL_ARG)
				fatal("This LLnextgen binary does not support the %.*s option. See the documentation for details\n", (int) optlength, CURRENT_OPTION);
			END_OPTION
			LONG_OPTION("dump-tokens", OPTIONAL_ARG)
				fatal("This LLnextgen binary does not support the %.*s option. See the documentation for details\n", (int) optlength, CURRENT_OPTION);
			END_OPTION
#endif
		}
		fatal("Option %.*s does not exist\n", (int) optlength, CURRENT_OPTION);
	NO_OPTION
		listAppend(option.inputFileList, safeStrdup(CURRENT_OPTION, "parseCmdLine"));
	END_OPTIONS
	
	if (option.dumpLexerwrapper) {
		fputs(option.threadSafe ? "int lexerWrapper(struct <prefix>this LLthis)\n" : "int lexerWrapper(void)\n", stdout);
		printf(lexerWrapperString, "<lexer>", option.threadSafe ? "LLthis" : "");
		fputs("\nWhere <lexer> should be replaced with the name of the lexical analyzer routine", stdout);
		if (option.threadSafe)
			fputs(",\n and <prefix> by LL or the prefix selected with %%prefix", stdout);
		fputs(".\nDon't forget to add \"%%lexical lexerWrapper;\" to your grammar!\n", stdout);
	}

	if (option.dumpLLmessage) {
		fputs(option.threadSafe ? "void <prefix>message(struct <prefix>this LLthis, int LLtoken)\n" : "void <prefix>message(int LLtoken)\n", stdout);
		fputs(llmessageString[0], stdout);
		fputs(llmessageString[1], stdout);
		fputs("\nWhere <prefix> shoult be replaced by LL or the prefix selected with %%prefix", stdout);		
	}
	
	if (option.dumpLexerwrapper || option.dumpLLmessage)
		exit(EXIT_SUCCESS);

	postOptionChecks();
	if (option.LLgenStyleOutputs && listSize(option.inputFileList) == 0)
		fatal("Cannot generate output from standard input %s\n", option.llgenMode ? "in LLgen mode" : "with --llgen-output-style");

	if (option.depend && option.dependCpp)
		fatal("Cannot generate dependency information and output C-preprocessor code at the same time\n");
	
	if (listSize(option.inputFileList) == 0 && option.outputBaseName == NULL)
		option.outputBaseName = "LLparser";
END_FUNCTION

/** Check the consistency of the options supplied by the user.
	
	This functions makes sure that the options supplied by the user are
	consistent, and either emits a warning or aborts with an error. Errors are
	only issued when files would be generated in a different place if the
	options had been consistent.
*/
void postOptionChecks(void) {
	if (option.llgenMode && option.verbose)
		fprintf(stderr, "This is LLnextgen, running as LLgen. Turning on compatibility options.\n");	

	if (option.generateLexerWrapper && option.noLLreissue) {
		fprintf(stderr, "warning: Option --no-llreissue cannot be used with --generate-lexer-wrapper\n");
		option.noLLreissue = false;
	}

	/* This can be done, but gives a hassle in generateLLgenStyle.c. As LLgen
	   can't do it either, we don't allow it. Note that simply ignoring these
	   options would result in different files being generated, and therefore
	   errors instead of warnings are used. */
	if (option.LLgenStyleOutputs && option.outputBaseNameSet)
		fatal("Cannot use --base-name %s\n", option.llgenMode ? "in LLgen mode" : "with --llgen-output-style");
	if (option.LLgenStyleOutputs && option.keepDirInFilename)
		fatal("Cannot use --keep-dir %s\n", option.llgenMode ? "in LLgen mode" : "with --llgen-output-style");

	if (option.threadSafe && option.reentrant)
		fatal("Cannot use --reentrant and --thread-safe at the same time\n");
	
	if (option.threadSafe && option.LLgenStyleOutputs)
		fatal("Cannot use --llgen-output-style with --thread-safe\n");

	if (option.threadSafe && option.noArgCount)
		fatal("Cannot use --no-arg-count for thread-safe parsers.\n");
	
	if (option.generateLLmessage)
		option.generateSymbolTable = true;
	
	if (option.warningsErrors && option.suppressWarnings) {
		static bool warned = false; /* Needed to suppress duplicate warnings */
		if (!warned) {
			fprintf(stderr, "warning: Using both --warnings-as-errors and --suppress-warnings makes no sense\n");
			warned = true;
		}
	}
}

/** Check the arguments for the --suppress-warnings option.
	@param optarg The arguments supplied for the option.
	@param optlength The length of @a current_option.
	@param current_option The current option (to allow correct printing of -W and --suppress-warnings)
*/
void checkWarningSuppressionArguments(const char *optarg, int optlength, const char *current_option) {
	if (optarg == NULL) {
		option.suppressWarnings = true;
	} else {
		char *subarg, *strtokState;
		/* Make a copy, for those C implementations in which argv[*] is not
		   writable */
		char *optargCopy = safeStrdup(optarg, "checkWarningSuppressionArguments");
		subarg = strtokReentrant(optargCopy, ",", &strtokState);
		if (subarg == NULL) {
			fprintf(stderr, "warning: argument for option %.*s is empty\n", optlength, current_option);
		}
		while (subarg != NULL) {
			if (strcmp(subarg, "arg-separator") == 0) {
				option.suppressWarningsTypes |= WARNING_ARG_SEPARATOR;
			} else if (strcmp(subarg, "option-override") == 0) {
				option.suppressWarningsTypes |= WARNING_OPTION_OVERRIDE;
			} else if (strcmp(subarg, "unbalanced-c") == 0) {
				option.suppressWarningsTypes |= WARNING_UNBALANCED_C;
			} else if (strcmp(subarg, "multiple-parser") == 0) {
				option.suppressWarningsTypes |= WARNING_MULTIPLE_PARSER;
			} else if (strcmp(subarg, "eofile") == 0) {
				option.suppressWarningsTypes |= WARNING_EOFILE;
			} else if (strcmp(subarg, "unused") == 0) {
				option.suppressWarningsTypes |= WARNING_UNUSED;
			} else if (strcmp(subarg, "reentrant-wrapper") == 0) {
				fprintf(stderr, "The reentrant-wrapper suppression is deprecated\n");
			} else if (strcmp(subarg, "datatype") == 0) {
				option.suppressWarningsTypes |= WARNING_DATATYPE;
			} else if (strncmp(subarg, "unused:", 7) == 0) {
				if (strlen(subarg) > 7) {
					if (option.unusedIdentifiers == NULL)
						option.unusedIdentifiers = newList();
					listAppend(option.unusedIdentifiers, safeStrdup(subarg + 7, "checkWarningSuppressionArguments"));
				} else {
					fprintf(stderr, "warning: argument 'unused:' for option %.*s needs an identifier\n", optlength, current_option);
				}
			} else {
				fatal("Argument '%s' is invalid for option %.*s\n", subarg, optlength, current_option);
			}
			subarg = strtokReentrant(NULL, ",", &strtokState);
		}
		free(optargCopy);
	}
}

/** Check the arguments for the --depend option.
	@param optarg The arguments supplied for the option.
*/
static void checkDependArguments(const char *optarg) {
	if (optarg == NULL) {
		option.suppressWarnings = true;
	} else {
		char *subarg, *strtokState;
		/* Make a copy, for those C implementations in which argv[*] is not
		   writable */
		char *optargCopy = safeStrdup(optarg, "checkDependArguments");
		subarg = strtokReentrant(optargCopy, ",", &strtokState);
		if (subarg == NULL) {
			fprintf(stderr, "warning: argument for option --depend is empty\n");
		}
		while (subarg != NULL) {
			if (strncmp(subarg, "targets:", 8) == 0) {
				if (strlen(subarg) > 7) {
					option.dependTargets = safeStrdup(subarg + 8, "checkDependArguments");
				} else {
					fprintf(stderr, "warning: argument 'targets:' for option --depend needs text\n");
				}
			} else if (strncmp(subarg, "file:", 5) == 0) {
				if (strlen(subarg) > 7) {
					option.dependFile = safeStrdup(subarg + 5, "checkDependArguments");
				} else {
					fprintf(stderr, "warning: argument 'file:' for option --depend needs file name\n");
				}
			} else if (strncmp(subarg, "extra-targets:", 14) == 0) {
				if (strlen(subarg) > 7) {
					option.dependExtraTargets = safeStrdup(subarg + 14, "checkDependArguments");
				} else {
					fprintf(stderr, "warning: argument 'extra-targets:' for option --depend needs text\n");
				}
			} else {
				fatal("Argument '%s' is invalid for option --depend\n", subarg);
			}
			subarg = strtokReentrant(NULL, ",", &strtokState);
		}
		free(optargCopy);
	}
}
