/* Copyright (C) 2005,2006 G.P. Halkes
   Licensed under the Open Software License version 2.0 */

#ifndef RULE_H
#define RULE_H

typedef struct NonTerminal NonTerminal;
typedef struct Alternative Alternative;

#include "nonRuleAnalysis.h"
#include "list.h"
#include "set.h"

#include "posixregex.h"

#define THREAD_NONTERMINAL -1

typedef struct {
	int flags;
	List *rule; /* List of Alternatives pointers */
	set first, follow, contains;
	int length;
	
	/* Members for tracing conflicts in the tree */
	int index;		/* Index in the alternative or THREAD_NONTERMINAL */
	union {
		Alternative *alternative;
		NonTerminal *nonTerminal;
	} enclosing;
} Term;

/* Note: these are numbered from 15 down because they share space with the
   TERM_XXX flags, and we want to be able to compile on 16 bit machines. */
#define NONTERMINAL_START (1<<15)
		/* This NonTerminal is mentioned in a %start directive. */
#define NONTERMINAL_REACHABLE (1<<14)
		/* Do we need to generate code for this non-terminal? */
#define NONTERMINAL_RECURSIVE_DEFAULT (1<<13)
		/* Is the default production recursive? */
#define NONTERMINAL_GLOBAL (1<<12)
		/* Is the non-terminal referenced from outside the file it was defined
		   in. This should only be calculated if multiple files are to be
		   generated. */
#define NONTERMINAL_VISITED (1<<11)
#define NONTERMINAL_TRACING (1<<10)
		/* Flags set by the trace routines to prevent recursion in the rules
		   inducing an infinite recursion in LLnextgen. Visited is set when
		   tracing into an alternative, traced is set when tracing calls to
		   the non-terminal. */
#define NONTERMINAL_TRACED (1<<9)
		/* Flag set by the trace routines to prevent multiple printing of the
		   same sub rules in conflicts. This mainly occurs in expression
		   constructs where precedence is implemented by splitting out the
		   different precedence levels into separate rules. */
#define NONTERMINAL_CONTAINS_CONFLICT (1<<8)
		/* Flag to indicate that rule must be printed for verbosity level 2 */
#define NONTERMINAL_FIRST (1<<7)
		/* This NonTerminal is mentioned in a %first directive, and therefore
		   unused non-terminal messages should be suppressed. */
/* WARNING: DON'T DEFINE ANYMORE FLAGS, AS THEY WILL CLASH WITH THE TERM_XXX FLAGS */

struct NonTerminal {
	Term term;
	CCode *parameters, *declarations;
	int argCount;
	Token *token;
	int number;
};

NonTerminal *newNonTerminal(Token *token, CCode *parameters, CCode *localDeclarations, List *rule, int argCount);

#define ALTERNATIVE_IF (1<<0)
#define ALTERNATIVE_DEFAULT (1<<1)
#define ALTERNATIVE_PREFER (1<<2)
#define ALTERNATIVE_AVOID (1<<3)
#define ALTERNATIVE_CONDITION (ALTERNATIVE_IF | ALTERNATIVE_PREFER | ALTERNATIVE_AVOID)
#define ALTERNATIVE_ACONFLICT (1<<4)
#define ALTERNATIVE_NULLABLE (1<<5)
#define ALTERNATIVE_NEVER_CHOSEN (1<<6)
#define ALTERNATIVE_NO_FINAL_READ (1<<7) 
	/* Flag IS NOT and SHOULD NOT be used for code generation, only to compute
	   the TERM_NO_FINAL_READ flag! */

struct Alternative {
	int flags;
	List *parts; /* List of GrammarPart pointers */
	Token *token;
	CCode *expression; /* Expression for %if */
	set first;
	set contains;
	int length;
	
	/* Label to jump to, to reach the code for this alternative. */
	int label;				

	Term *enclosing;
#ifdef MEM_DEBUG
	bool dontFreeToken;
#endif
};

Alternative *newAlternative(void);

#define TERM_PERSISTENT (1<<0)
#define TERM_WHILE (1<<1)
#define TERM_NULLABLE (1<<2)	/* Set ONLY if the term itself can produce NULL, not when
									it is followed by a nullable repetition operator */
#define TERM_OPTMULTREP (1<<3)	/* Set if a term can optionally repeat more than once
									For non-terminals this means that in SOME rule they
									can optionally repeat more than once. */
#define TERM_MULTIPLE_ALTERNATIVE (1<<4)
#define TERM_RCONFLICT (1<<5)
#define TERM_NO_FINAL_READ (1<<6)
/* WARNING: DO NOT DEFINE ANY MORE FLAGS, AS THIS WILL CONFLICT WITH THE FLAGS
   DEFINED FOR NON-TERMINALS!!! */

/* Parts are clasified as a bitfield to make PART_IS_TERMINAL evaluate
	its argument only once. Other implementations of PART_IS_TERMINAL
	that only evalute their argument once are less portible (i.e. uses typeof).
*/
typedef enum {
	PART_UNKNOWN = (1<<0),		/* Part created by parser error handling. */
	PART_ACTION = (1<<1),
	PART_NONTERMINAL = (1<<2),
	PART_TERMINAL = (1<<3),		/* %token defined */
	PART_LITERAL = (1<<4),		/* character literal */
	PART_UNDETERMINED = (1<<5),	/* Token or non-terminal, lookup has to determine */
	PART_TERM = (1<<6)			/* Something between [] */
} GrammarPartSubtype;
#define PART_IS_TERMINAL(x) ((*(x)).subtype & (PART_LITERAL | PART_TERMINAL))

/* Repeats are clasified as a bitfield to make REPEATS_NULLABLE evaluate
	its argument only once
*/
typedef enum {
	FIXED = (1<<0),
	STAR = (1<<1),
	PLUS = (1<<2)
} RepeatsSubtype;
#define REPEATS_NULLABLE(x) ((x)->repeats.subtype & STAR)

typedef struct {
	RepeatsSubtype subtype;
	int number;	/* -1 for none supplied */
} Repeats;

typedef struct {
	GrammarPartSubtype subtype;
	Token *token;
	Repeats repeats; /* FIXME: doesn't belong here, really. It belongs in Term */
	union {
		struct {
			Term term;
			CCode *expression;	/* %while expression */
		} term;
		
		struct {
			NonTerminal *nonTerminal;	/* The non-terminal that is called */
			CCode *expression;			/* Parameters */
		} nonTerminal;
		
		CCode *action;		/* Action */
		
		struct {
			int terminal;		/* Terminal */
		} terminal;
	} un;
#ifdef MEM_DEBUG
	bool dontFreeToken;
#endif
} GrammarPart;

GrammarPart *newGrammarPart(Token *token, GrammarPartSubtype subtype);
GrammarPart *wrapInTerm(GrammarPart *grammarPart);

void setOnlyReachable(bool newValue);

void determineNullability(void);
void computeFirstSets(void);
void computeFollowSets(void);
void computeLengths(void);
void computeContainsSets(void);
void computeFinalReads(void);
void determineReachabilitySimple(void);
#ifdef USE_REGEX
void scanForUnidentifiedTokens(void);
#endif
void resolve(NonTerminal *nonTerminal);
void allocateSets(void);
void fixupFirst(void);
void setAttributes(void);
void findConflicts(void);
void setDefaults(void);
void enumerateSets(void);

void walkRules(void (*nonTerminalAction)(NonTerminal *));
void walkRulesSimple(void (*termAction)(Term *));
#endif
